// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file AdaptBase.h */

#ifndef AMDIS_ADAPTBASE_H
#define AMDIS_ADAPTBASE_H

#include <string>

namespace AMDiS {

  class ProblemIterationInterface;
  class ProblemTimeInterface;
  class AdaptInfo;

  // ============================================================================
  // ===== class AdaptBase ======================================================
  // ============================================================================

  /** \brief
   * Interface for adaption loops.
   */
  class AdaptBase
  {
  public:
    /** \brief
     * Constructor
     */
    AdaptBase(const ::std::string&         name,
	      ProblemIterationInterface *problemIteration,
	      AdaptInfo                 *adaptInfo,
	      ProblemTimeInterface      *problemTime = NULL,
	      AdaptInfo                 *initialAdaptInfo = NULL)
      : name_(name),
	problemIteration_(problemIteration),
	adaptInfo_(adaptInfo),
	problemTime_(problemTime),
	initialAdaptInfo_(initialAdaptInfo)
    {};

    /** \brief
     * Destructor
     */
    virtual ~AdaptBase() {};

    /** \brief
     * Pure virtual method. Must be overloaded by sub classes to perform
     * a concrete adaption loop. 
     */
    virtual int adapt() = 0;

    /** \brief
     * Returns \ref name_
     */
    inline const ::std::string& getName() const { 
      return name_; 
    };

    /** \brief
     * Returns \ref problemIteration_
     */
    inline ProblemIterationInterface *getProblemIteration() {
      return problemIteration_;
    };

    inline void setProblemIteration(ProblemIterationInterface *pii) {
      problemIteration_ = pii;
    };

    /** \brief
     * Returns \ref adaptInfo_
     */
    inline AdaptInfo *getAdaptInfo() { 
      return adaptInfo_; 
    };

    /** \brief
     * Returns \ref problemTime_
     */
    inline ProblemTimeInterface *getProblemTime() {
      return problemTime_;
    };

    inline void setProblemTime(ProblemTimeInterface *pti) {
      problemTime_ = pti;
    }

    /** \brief
     * Returns \ref initialAdaptInfo_
     */
    inline AdaptInfo *getInitialAdaptInfo() { 
      return initialAdaptInfo_; 
    };

  protected:
    /** \brief
     * Name of the adaption loop
     */
    ::std::string name_;

    /** \brief
     * Problem iteration interface
     */
    ProblemIterationInterface *problemIteration_;

    /** \brief
     * Main adapt info
     */
    AdaptInfo *adaptInfo_;

    /** \brief
     * problem time interface
     */
    ProblemTimeInterface *problemTime_;

    /** \brief
     * Adapt info for initial adapt. Will be given to 
     * problemTime_->solveInitialProblem().
     */
    AdaptInfo *initialAdaptInfo_;

    /** \brief
     * Info level
     */
    static int info_;
  };

}

#endif
