// ============================================================================
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file GMResSolver.h */

#ifndef AMDIS_GMRESSOLVER_H
#define AMDIS_GMRESSOLVER_H

#include <vector>
#include "OEMSolver.h"
#include "MemoryManager.h"

namespace AMDiS {

  // ============================================================================
  // ===== class GMResSolver ===================================================
  // ============================================================================

  /** 
   * \ingroup Solver
   *
   * \brief
   * Solves a linear system by the restarted GMRES(m) method (Generalized Minimal
   * Residual) for general nonsymmetric system matrices.
   *
   * The implementation is based on the following book: "Numerik linearer
   * Gleichungssystene", 2. Auflage, Andreas Meister. The algorithm is described
   * on pages 150 and  154 (for the restarted version). The extension to the
   * preconditioned GMRES(m) is based on the description in the book "Iterative
   * Methods for Sparse Linear Systems", second edition, Yousef Saad, on page 268.
   *
   * The orthogonal system is build using the modified Gram Schmidt (MGS) method.
   * The least square problem is solved using Givens transformations.
   * 
   */
  template<typename VectorType>
  class GMResSolver : public OEMSolver<VectorType>
  {
  public:
    MEMORY_MANAGED(GMResSolver<VectorType>);

    /** \brief
     * Creator class used in the OEMSolverMap.
     */
    class Creator : public OEMSolverCreator<VectorType>
    {
    public:
      MEMORY_MANAGED(Creator);

      virtual ~Creator() {};

      /** \brief
       * Returns a new GMResSolver object.
       */   
      OEMSolver<VectorType>* create() { 
	return NEW GMResSolver<VectorType>(this->name);
      };
    };

    /** \brief
     * constructor
     */
    GMResSolver(::std::string name);

    /** \brief
     * destructor
     */
    ~GMResSolver();

  protected:
    /** \brief
     * realisation of OEMSolver::solveSystem
     */
    int solveSystem(MatVecMultiplier<VectorType> *mv, VectorType *x, VectorType *b);

    /** \brief
     * realisation of OEMSolver::init
     */
    void init();

    /** \brief
     * realisation of OEMSolver::exit
     */
    void exit();

    /** \brief
     * One run of the GMRES(m) method. Is called from solveSystem().
     */
    int gmres(MatVecMultiplier<VectorType> *mv,
	      VectorType *x, 
	      VectorType *b);

  private:
    /** \brief
     * Stores the tolerance boundary for numerical computations.
     */
    double TOL_;

    /** \brief
     * The parameter m of GMRES(m), i.e. the number of iterations before a restart of
     * the algorithm is forced.
     */
    int restart_;

    /** \brief
     * Stores intermediate results in the orthogonalization step.
     */
    VectorType *w_;


    VectorType *z_;

    VectorType *r_;

    /** \brief
     * Pointers to the vectors of the orthogonal system.
     */
    VectorType **v;

    /** \brief
     * Stores the gamma values for Givens transformations.
     */
    ::std::vector<double> gamma;

    /** \brief
     * Stores the c values for Givens transformations.
     */    
    ::std::vector<double> c;

    /** \brief
     * Stores the s values for Givens transformations.
     */
    ::std::vector<double> s;

    ::std::vector<double> y_;

    /** \brief
     * H-Matrix computed in the GMRES algorithm.
     */    
    ::std::vector< ::std::vector<double> > h;
  };


}

#include "GMResSolver.hh"

#endif // AMDIS_GMRESSOLVER_H
